#include "ofnfs.h"

int
ofnfs_read_vnode(void *_vol, vnode_id vnid, char reenter, void **_node)
{
	int err;
	nspace	*vol = (nspace*)_vol;

	if (vol == NULL) {
		dprintf("ofnfs : read_vnode given bad vol\n");
		return EINVAL;
	}

	dprintf("ofnfs : read_vnode called with vnid %Lx\n", vnid);
	
	// "There can be only one."  Because, read_vnode won't get
	// called on the root point, ever - the fsil already knows about
	// it and has it loaded.  So, this is either being called on our
	// one file, or the fsil is talking trash.
	if (vnid == root_vnid) {
		*_node = vol->root_vnode;
		return B_OK;
	} else if (vnid == the_one) {
		*_node = vol->file_vnode;
		return B_OK;
	}
	
	return ENOENT;
}


int
ofnfs_write_vnode(void *_vol, void *_node, char reenter)
{
	int	err;
	nspace	*vol = (nspace*)_vol;
	vnode	*node = (vnode*)_node;

	// Normally, this is called to let your filesystem know it is
	// done (for the moment) with a node, so you would free up any associated
	// memory or other resources.  (It doesn't imply that anything should
	// be written to disk, or across the network.)
	// This is really easy for ofnfs, since handle our two vnodes with mount and unount.
	return B_OK;
}


int
ofnfs_access(void *_vol, void *_node, int mode) {
	
	int err;
	nspace	*vol = (nspace*)_vol;
	vnode	*node = (vnode*)_node;
	
	return B_OK;
}



int
ofnfs_walk(void *_vol, void *_dir, const char *file, char **newpath, vnode_id *_vnid)
{
	int err;
	nspace *vol = (nspace*)_vol;
	vnode *dir = (vnode*)_dir;
	vnode *vn = NULL;
	

	dprintf("ofnfs : nsid %d walk %s %Lx\n", vol->nsid, file, dir->vnid);

	if ( (vol == NULL) || (dir == NULL))
		return EINVAL;

	// If we're not called with the root vnode, then we're
	// being called on a non-directory vnode.
	if (dir->vnid != root_vnid)
		return ENOENT;

	if (strcmp(file, ".") == 0) {
		*_vnid = root_vnid;
		dprintf("calling get_vnode : %d %Lx\n", vol->nsid, root_vnid);
		err = get_vnode(vol->nsid,root_vnid,(void**)&vn);
		if (err != B_OK)
			dprintf("ofnfs : walk returning %s first\n", strerror(err));	
		return err;
	}

	// We wont get asked about ".." cause we dont know who that is; the
	// fsil handles that kind of glue between mount points.
	
	// Now we need to see if we have this filename.
	err = get_vnode(vol->nsid,the_one,(void**)&vn);
	if (err != B_OK) {
		dprintf("enoent in walk\n");
		return ENOENT;
	}	
	LOCK(vn->lock);
	if (strcmp(file, vn->filename) == 0)
		err = B_OK;
	else
		err = ENOENT;
	UNLOCK(vn->lock);
	
	if (err == B_OK)
		*_vnid = the_one;
	else
		put_vnode(vol->nsid,the_one);
		

	if (err != B_OK)
		dprintf("ofnfs : walk returning %s\n", strerror(err));	
	return err;
}




int
ofnfs_opendir(void *_vol, void *_node, void **_cookie)
{
	int err;
	nspace *vol = (nspace*)_vol;
	vnode *node = (vnode*) _node;
	dircookie *cookie;
	
	if (node->vnid != root_vnid)
		return ENOTDIR;
	
	cookie = (dircookie*)malloc(sizeof(dircookie));
	if (cookie == NULL) return ENOMEM;

	dprintf("ofnfs : opendir on nsid %d dir %Lx\n", vol->nsid, node->vnid);	
	cookie->index = -2;
	*_cookie = cookie;
	return B_OK;
}


int
ofnfs_readdir(void *_vol, void *_dir, void *_cookie, long *num,
					struct dirent *entry, size_t bufsize)
{
// Opendir sets the cookie index to -2.  We'll increment, and get that entry's info.
	int			err;
	nspace*		vol = (nspace*)_vol;
	vnode*		dir = (vnode*)_dir;
	dircookie*	cookie = (dircookie*)_cookie;
	vnode*		vn = NULL;
	
	if (cookie == NULL)
		return EINVAL;

	dprintf("ofnfs : readdir on nsid %d, dir %Lx, cookie %d\n", vol->nsid,
		dir->vnid, cookie->index);
		
	if (cookie->index == -2) {
		strcpy(entry->d_name, ".");
		entry->d_ino = dir->vnid;
		entry->d_reclen = 2; // this is strlen(".");
		entry->d_dev = vol->nsid;
		*num = 1;
		cookie->index++;
		dprintf("ofnfs : readdir returning vnid %x %s\n",entry->d_ino, entry->d_name);
		return B_OK;
	} else if (cookie->index == -1) {
		// We won't get asked about
		strcpy(entry->d_name, "..");
		entry->d_ino = dir->vnid;
		entry->d_reclen = 2; // this is strlen(".");
		entry->d_dev = vol->nsid;
		*num = 1;
		cookie->index++;
		dprintf("ofnfs : readdir returning vnid %x %s\n",entry->d_ino, entry->d_name);
		return B_OK;
	} else if (cookie->index > 0) {
		// They want more!  But we only have one file.
		// This isn't an error, just say that *num = 0.
		*num = 0;
		dprintf("ofnfs : readdir returning 0 entries\n");
		return B_OK;
	}
	
	// So, cookie->index is 0, so whats the current name of
	// our file?
	err = get_vnode(vol->nsid,the_one,(void**)&vn);
	if (err != B_OK) {
		*num = 0;
		dprintf("ofnfs : readdir returning 0 entries cause get_vnode failed\n");
		return B_OK;
	}
	
	LOCK(vn->lock);
	memcpy(entry->d_name, vn->filename, B_FILE_NAME_LENGTH);
	UNLOCK(vn->lock);
	put_vnode(vol->nsid,the_one);
	
	entry->d_ino = the_one;
	entry->d_reclen = strlen(entry->d_name) + 1;
	entry->d_dev = vol->nsid;
	*num = 1;
	cookie->index++;
	dprintf("ofnfs : readdir : THE vnid %x %s\n",entry->d_ino, entry->d_name);
	return B_OK;
}


int
ofnfs_rewinddir(void *_vol, void* _node, void* _cookie)
{
	nspace* 	vol = _vol;
	vnode*		vnode = _node;
	dircookie*	cookie = (dircookie*)_cookie;

	if (cookie == NULL)
		return EINVAL;

	dprintf("ofnfs : rewinddir\n");
		
	cookie->index = -2;
	return B_OK;
}


int
ofnfs_free_dircookie(void* _vol, void* _node, void* _cookie)
{
	nspace*		vol = (nspace*)_vol;
	vnode*		node = (vnode*)_node;
	dircookie*	cookie = (dircookie*)_cookie;
	
	if (cookie == NULL)
		return EINVAL;

	dprintf("ofnfs : free_dircookie\n");
	
	free(cookie);
	return B_OK;
}



int
ofnfs_closedir(void *_vol, void *_node, void *_cookie)
{
	dprintf("ofnfs : closedir\n");
	return B_NO_ERROR;
}


int	
ofnfs_open(void* _vol, void* _node, int omode, void** _cookie)
{
	status_t	err;
	nspace*		vol = (nspace*)_vol;
	vnode*		node = (vnode*)_node;
	filecookie	*cookie = NULL;

	cookie = (filecookie*)malloc(sizeof(filecookie));
	if (cookie == NULL)
		return ENOMEM;

	dprintf("ofnfs : open\n");

	
	*_cookie = cookie;
	return B_OK;
}


int
ofnfs_close(void *_vol, void *_node, void *_cookie)
{
	status_t err;
	nspace *vol = (nspace *)_vol;
	vnode *node = (vnode *)_node;
	filecookie	*cookie = _cookie;
	dprintf("ofnfs : close\n");

	return B_OK;
}


int
ofnfs_free_cookie(void *_vol, void *_node, void *_cookie)
{
	status_t err;
	nspace *vol = (nspace *)_vol;
	vnode *node = (vnode *)_node;
	filecookie *cookie = (filecookie*)_cookie;
		
	if (cookie == NULL)
		return EINVAL;

	dprintf("ofnfs : free_cookie\n");
	
	free(cookie);

	return B_OK;
}



int
ofnfs_read(void *_vol, void *_node, void *_cookie, off_t pos,
				void *buf, size_t *len)
{
	int err;				
	nspace		*vol = (nspace*)_vol;
	vnode		*node = (vnode*)_node;
	filecookie		*cookie = (filecookie*)_cookie;
	uint32		cur_time;

	*len = 0;
	return B_OK;
}


int
ofnfs_rstat(void* _vol, void* _node, struct stat* st)
{
	nspace*		vol = (nspace*)_vol;
	vnode*		node = (vnode*)_node;

	if ( (vol==NULL) || (node==NULL))
		return EINVAL;	
	
	st->st_dev = vol->nsid;

	st->st_ino = node->vnid;
	st->st_mode = (S_IRUSR |  S_IRGRP | S_IROTH);
	if (node->vnid == the_one)
		st->st_mode |= S_IFREG;
	else
		st->st_mode |= S_IFDIR;

	st->st_nlink = 1;
	st->st_uid = 0;
	st->st_gid = 0;
	st->st_size = 0;
	st->st_blksize = 1024;
	
	LOCK(node->lock);
	st->st_atime = node->last_rename_time;
	st->st_ctime = node->last_rename_time;
	st->st_mtime = node->last_rename_time;
	st->st_crtime = node->last_rename_time;
	UNLOCK(node->lock);

	dprintf("ofnfs : rstat\n");

	return B_OK;	
}

int form_vnodes(nspace *vol);
int clear_vnodes(nspace *vol);


int
ofnfs_mount(nspace_id nsid, const char* device, ulong flags, void *_parms,
				size_t len, void **data, vnode_id *vnid)
{
	int err;
	parms_t*	parms = (parms_t*)_parms;
	nspace*		vol = NULL;
	vnode_id 	this_vnid;

	if ( (device != NULL) || (parms == NULL)) {
		err = B_ERROR;
		goto exit1;
	}
	
	vol = (nspace*)malloc(sizeof(struct nspace));
	if (vol==NULL) {
		err = ENOMEM;
		goto exit1;
	}

	// Copy parms;
	if (len != sizeof(parms_t)) {
		err = EINVAL;
		goto exit2;
	}
	memcpy(& vol->params, parms, sizeof(parms_t));
	
	// Form the two vnodes for the file system
	err = form_vnodes(vol);
	if (err != B_OK) goto exit2;

	//Start up our network
	err = begin_network(vol);
	if (err != B_OK) goto exit3;


	// This way, you can see stack traces with the
	// kernel debugger.
	load_driver_symbols("ofnfs");
	
	// Report back to the fsil our stuff
	vol->nsid = nsid;
	*vnid = root_vnid;
	*data = vol;

	// Tell the fsil there's a new vnode in town.
	err = new_vnode(nsid,root_vnid,vol->root_vnode);
	if (err != B_OK) goto exit3;

	// And we're done.
	dprintf("ofnfs : nsid %x root_vnid %x\n", vol->nsid, *vnid);
	return B_OK;
	

	// Now the error exits.
exit4:
	end_network(vol);
exit3:
	clear_vnodes(vol);
exit2:
	free(vol);
exit1:
	dprintf("ofnfs : mount returning error %s\n", strerror(err));
	return err;
}

int
ofnfs_unmount(void *_vol)
{
	int err;
	nspace *vol = (nspace*)_vol;

	// Bring everything down
	end_network(vol);
	clear_vnodes(vol);
	free(vol);
	return B_OK;
}


int
ofnfs_rfsstat(void *_vol, struct fs_info *fss)
{
	nspace*		vol = (nspace*)_vol;

	// Just made up...
	fss->block_size = 1024;
	fss->total_blocks = 1;
	fss->free_blocks = 0;

	// fss->dev and fss->root filled in by kernel
	fss->flags = B_FS_IS_SHARED | B_FS_IS_PERSISTENT;
	// IO size - specifies buffer size for file copying
	fss->io_size = 65536;
	// Device name.
	strcpy(fss->device_name, "");
	// Volume name.
	strcpy(fss->volume_name, "A truly trivial netfs");
	// File system name
	strcpy(fss->fsh_name, "ofnfs");
	
	return B_OK;
}



int
form_vnodes(nspace *vol)
{
	int err;
	char name[32];

	vol->root_vnode = (vnode*)malloc(sizeof(vnode));
	if (vol->root_vnode == NULL)
		return ENOMEM;
	
	vol->file_vnode = (vnode*)malloc(sizeof(vnode));
	if (vol->file_vnode == NULL) {
		free(vol->root_vnode);
		return ENOMEM;
	}

	sprintf(name, "ofnfs root lock%x", vol->nsid);
	if ((err = new_lock(&(vol->root_vnode->lock), name)) != B_OK) {
		dprintf("couldnt create root vnode lock.\n");
		free(vol->root_vnode);
		free(vol->file_vnode);
		return err;
	}

	sprintf(name, "ofnfs file lock%x", vol->nsid);
	if ((err = new_lock(&(vol->file_vnode->lock), name)) != B_OK) {
		dprintf("couldnt create file vnode lock.\n");
		free(vol->root_vnode);
		free(vol->file_vnode);
		free_lock(&(vol->root_vnode->lock));
		return err;
	}

	memset(vol->root_vnode, 0, sizeof(vnode));
	memset(vol->file_vnode, 0, sizeof(vnode));

	vol->root_vnode->vnid = root_vnid;
	vol->file_vnode->vnid = the_one;
	vol->file_vnode->parent_vnid = root_vnid;
	
	strcpy(vol->file_vnode->filename, "No traffic yet");
	return B_OK;
}

int
clear_vnodes(nspace *vol)
{

	if (vol->root_vnode)
		free(vol->root_vnode);
	
	if (vol->file_vnode)
		free(vol->file_vnode);
	
	return B_OK;
}



vnode_ops fs_entry = {
	&ofnfs_read_vnode,
	&ofnfs_write_vnode,
	NULL, // remove_vnode
	NULL, // secure
	&ofnfs_walk,
	&ofnfs_access, // access
	NULL, // create 
	NULL, // mkdir
	NULL, // symlink
	NULL, // link
	NULL, // rename
	NULL, // unlink
	NULL, // rmdir
	NULL, // readlink
	&ofnfs_opendir,
	&ofnfs_closedir,
	&ofnfs_free_dircookie,
	&ofnfs_rewinddir,
	&ofnfs_readdir,
	&ofnfs_open,
	&ofnfs_close,
	&ofnfs_free_cookie,
	&ofnfs_read,
	NULL, // write
	NULL, // readv
	NULL, // writev
	NULL, // ioctl
	NULL, // setflags
	&ofnfs_rstat,
	NULL, // wstat
	NULL, // fsync
	NULL, // initialize
	&ofnfs_mount,
	&ofnfs_unmount,
	NULL, // sync
	&ofnfs_rfsstat,
	NULL, // wfsstat

	NULL, // select
	NULL, // deselect

	NULL, // open_indexdir
	NULL, // close_indexdir
	NULL, // free_indexdircookie
	NULL, // rewind_indexdir
	NULL, // read_indexdir

	NULL, // create_index
	NULL, // remove_index
	NULL, // rename_index
	NULL, // stat_index

	NULL, // open_attrdir
	NULL, // close_attrdir
	NULL, // free_attrdircookie
	NULL, // rewind_attrdir
	NULL, // read_attrdir

	NULL, // write_attr
	NULL, // read_attr
	NULL, // remove_attr
	NULL, // rename_attr
	NULL, // stat_attr

	NULL, // open_query
	NULL, // close_query
	NULL, // free_querycookie
	NULL  // read_query
};

int32	api_version = B_CUR_FS_API_VERSION;
