#include "ofnfs.h"
#include "ksocket.h"
#include <errno.h>


// Below are all the network functions.  You'll notice they're small and simple.
// The really good reason for this is that if something goes wrong, there's an easy
// and logical way to teardown our network connection.
//begin_network(nspace *vol);
// Here's the list in the order that they're called:
//init_ksocket();
//init_listening_socket(nspace *vol);
//spawn_listening_thread(nspace *vol);
//listener_thread(void *_vol);
//isDataPending(int socket, uint32 timeout);
//end_network(nspace *vol);
//kill_listening_thread(nspace *vol);
//kill_listening_socket(nspace *vol);
//kill_ksocket();


int
listener_thread(void *_vol)
{
	int err;
	nspace *vol = (nspace*)_vol;

	struct sockaddr from;
	int fromlen;	
	char buf[B_FILE_NAME_LENGTH];
	vnode *vn=NULL;
	
	while ((err = acquire_sem(vol->listener_sem) == B_OK)) {

		// Wait 5 seconds for a transmission, then reloop, so that we can be
		// unmounted.
		err = isDataPending(vol->sockid, 5);
		if (err == B_TIMED_OUT) {
			release_sem(vol->listener_sem);
			continue;
		}
		if (err == B_OK) {
			err = krecvfrom(vol->sockid, buf, B_FILE_NAME_LENGTH, 0, &from, &fromlen);
		}	

		if (err < 0) {
			dprintf("ofnfs:  err : %d %s |  errno %d %s\n", err,strerror(err), errno,strerror(errno));
			// Here's where things get interesting.  Now, a couple of things could have gone wrong;
			// Ksocketd could have died. The net_server could have died (or been restarted).
			// You have to decide what you want to do in the face of these dangers.  You could
			// just return from this function, and you wont be able to listen for broadcasts anymore.
			// _Or_, you could try and restart your network connection.  If you do that, you'll probably
			// want to enforce some kind of restart policy.  (Do you want to restart everytime?  Do you want
			// to pause between restarts?  Do you want to restart a maximum number of times?  If you know the
			// net connection is bad, what do you do with further requests for files? ...)
			
			// Hmm, lets give networking a swift kick.
			end_network(vol);
			err = begin_network(vol);
			if (err != B_OK) {
				// Well, that didn't help.  I give up.  (Although you don't have to; what's your network failure policy?)
				dprintf("ofnfs: listener thread giving up and dying after failed restart\n");
				release_sem(vol->listener_sem);
				return B_ERROR;
			} else {
				// Ah good!  Let's try this again.
				release_sem(vol->listener_sem);
				continue;
			}
		} else {
			// Life is good sometimes.
			dprintf("ofnfs: got %d size packet\n", err);
			err = get_vnode(vol->nsid,the_one,(void**)&vn);
			if (err != B_OK) {
				release_sem(vol->listener_sem);
				continue;
			}
			LOCK(vn->lock);
			strncpy(vn->filename, buf, B_FILE_NAME_LENGTH);
			// Ah, strncpy.  Good, good.  But what are we not doing here?  We're not checking
			// for bad file names.  Do you really want a file called /////// ?  Probably not. 
			vn->last_rename_time=real_time_clock();
			//Now notify listeners that the name has changed.
			//notify_listener(B_ENTRY_MOVED, <name space id>, <old dir vnid>, <new dir vnid>, <node's vnid>, <new file name>)
			notify_listener(B_ENTRY_MOVED, vol->nsid, root_vnid, root_vnid, the_one, vn->filename);
			//Also report that the file statistics have changed.
			notify_listener(B_STAT_CHANGED, vol->nsid, 0,0,the_one,NULL);
			UNLOCK(vn->lock);
			put_vnode(vol->nsid,the_one);
			release_sem(vol->listener_sem);
			continue;
		}
	}
	
	dprintf("ofnfs: listener thread exiting with err %s\n", strerror(err));
	return err;
}



// The atomic add's below ensure that even if your file system gets used twice
// (two different mount points) you only initialize the ksocket code once.
static int32 ofnfs_open_count = 0;

status_t
init_ksocket() {
	status_t err;
	if (atomic_add(&ofnfs_open_count, 1) > 0) {
		return B_OK;
    } else {
		dprintf("ofnfs: initing kernel sockets\n");
		err = ksocket_init();
		if (err != B_OK) {
			atomic_add(&ofnfs_open_count, -1);
			return err;
		} else {
			return B_OK;
		}
	}
}	

void
kill_ksocket() {
	
	if (atomic_add(&ofnfs_open_count, -1) > 1) {
		return;
	} else {
		dprintf("ofnfs: killing socket\n");
		ksocket_cleanup();
	}
}


int
init_listening_socket(nspace *vol)
{
	int err;
	size_t s;
	struct sockaddr_in sa;

	vol->sockid = ksocket(AF_INET, SOCK_DGRAM, 0);
	if (vol->sockid < 0) {
		dprintf("ofnfs : couldnt create socket\n");
		return errno;
	}
	sa.sin_family = AF_INET;
	sa.sin_port = htons(vol->params.port);
	sa.sin_addr.s_addr = htonl(INADDR_ANY);

	s = sizeof(sa);
	err = kbind(vol->sockid, (const struct sockaddr *) & sa, s);
	if (err != B_OK) {
		dprintf("ofnfs : couldnt bind socket\n");
		kclosesocket(vol->sockid);
		return errno;
	}

	return B_OK;
}	

int
kill_listening_socket(nspace *vol)
{
	return kclosesocket(vol->sockid);
}


int
spawn_listening_thread(nspace *vol)
{
	char name[32];
	sprintf(name, "ofnfs thread %d", system_time());
	if ((vol->listener_sem = create_sem(1,name)) < 0)
		goto exit1;
	
	if ((vol->listener_thread = spawn_kernel_thread(listener_thread, name, B_LOW_PRIORITY, (void*)vol)) < 0)
		goto exit2;
	
	if (resume_thread(vol->listener_thread) < 0)
		goto exit2;
	
	// Thread has been spawned, all is good.
	return B_OK;


exit2:	
	delete_sem(vol->listener_sem);
exit1:
	return B_ERROR;
}

int
kill_listening_thread(nspace *vol)
{
	int32 dummy;
	delete_sem(vol->listener_sem);
	wait_for_thread(vol->listener_thread, &dummy);
	return B_OK;
}

int
begin_network(nspace *vol)
{
	int err;
	char name[32];
	

	err = init_ksocket(vol);
	if (err != B_OK) return err;
		
	err = init_listening_socket(vol);
	if (err != B_OK) goto exit1;
		
	err = spawn_listening_thread(vol);
	if (err != B_OK) goto exit2;
	
	return B_OK;

exit2:
	kill_listening_socket(vol);
exit1:
	kill_ksocket();
	return B_ERROR;
}


int
end_network(nspace *vol)
{
	kill_listening_thread(vol);
	kill_listening_socket(vol);
	kill_ksocket(vol);
	return B_OK;
}



// This is how we make sure there's data waiting for us
// _before_ we call recvfrom.

int
isDataPending(int socket, uint32 timeout)
{
	int err;
	fd_set fd;
	struct timeval tv;

	if (timeout > 0)
		tv.tv_sec = timeout;
    
    FD_ZERO(&fd);
   	FD_SET(socket, &fd);
    	
	err = kselect(socket + 1, &fd, NULL, NULL, &tv);
	if (err == 1) {	
		if (FD_ISSET(socket, &fd))
		    return B_OK;  // Data is pending
	} else if (err == 0) {
		dprintf("select timed out\n");
		return B_TIMED_OUT;  // No data
   	} else {
		dprintf("exiting isDataPending result is %d %s\n",err, strerror(err));
		return B_IO_ERROR; // error
	}
	
	//not reached
	return B_ERROR;			
}










